"""Tests for metabot.modules.moderator."""

import pytest

from metabot.modules import moderator


@pytest.fixture
def conversation(build_conversation):  # pylint: disable=missing-docstring
    return build_conversation(moderator)


# pylint: disable=line-too-long


def test_forward(conversation):  # pylint: disable=redefined-outer-name
    """Test autoforward."""

    assert conversation.message('Test Forward', chat_type='channel') == ''

    conversation.bot.config['issue37']['moderator']['-1002000002000']['forward'][
        'from'] = -1001000001000

    assert conversation.message('Test Forward', chat_type='channel') == """\
[forward_message chat_id=-1002000002000 disable_notification=True from_chat_id=-1001000001000 message_id=2000]
(EMPTY MESSAGE)
"""


def test_mod(conversation):  # pylint: disable=redefined-outer-name
    """Test /mod."""

    modconf = conversation.bot.config['issue37']['moderator']
    groupdata = conversation.multibot.conf['groups']
    modconf['-1001000001000']['title'] = groupdata[-1001000001000]['title'] = 'Mod Test'
    modconf['-1002000002000']['title'] = groupdata[-1002000002000]['title'] = 'Hidden Group'

    assert conversation.message('/mod', user_id=2000) == """\
[chat_id=2000 disable_web_page_preview=True parse_mode=HTML]
Group Admin: <b>Become a group admin</b>

Hi! You aren't an admin in any groups I'm in. If you should be, ask a current admin to promote you from the group's members list.
"""

    conversation.multibot.conf['groups'][-1001000001000]['admins'] = [2000]

    assert conversation.message('/mod', user_id=2000) == """\
[chat_id=2000 disable_web_page_preview=True parse_mode=HTML]
Group Admin: <b>Choose a group</b>
[-1001000001000 • Mod Test | /mod -1001000001000]
"""

    assert conversation.message('/mod -1002000002000', user_id=2000) == """\
[chat_id=2000 disable_web_page_preview=True parse_mode=HTML]
Group Admin: <b>Choose a group</b>

I can't set <code>-1002000002000</code>.
[-1001000001000 • Mod Test | /mod -1001000001000]
"""

    assert conversation.message('/mod -1001000001000', user_id=2000) == """\
[chat_id=2000 disable_web_page_preview=True parse_mode=HTML]
Group Admin › -1001000001000: <b>Choose a field</b>
[calendars • Which calendars should be listed in /events? | /mod -1001000001000 calendars]
[daily • Should I announce upcoming events once a day? | /mod -1001000001000 daily]
[forward • Automatically forward messages from one chat to this one. | /mod -1001000001000 forward]
[greeting • How should I greet people when they join? | /mod -1001000001000 greeting]
[maxeventscount • How many events should be listed in /events? | /mod -1001000001000 maxeventscount]
[maxeventsdays • How many days into the future should /events look? | /mod -1001000001000 maxeventsdays]
[timezone • What time zone should be used in /events? | /mod -1001000001000 timezone]
[Back | /mod]
"""


def test_admin(conversation):  # pylint: disable=redefined-outer-name
    """Test /admin BOTNAME moderator."""

    assert conversation.message('/admin modulestestbot moderator') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator: <b>Add me to a group</b>

I'm not in any groups! Add me to an existing group from its details screen.
[Back | /admin modulestestbot]
"""

    adding_user = {'id': 2000}
    joined_user = {'id': 3000, 'is_bot': True, 'first_name': 'User 3000'}
    chat = {'id': -1001000001000, 'type': 'supergroup', 'title': 'My Group'}
    message = {
        'from': adding_user,
        'chat': chat,
        'message_id': 5000,
        'new_chat_members': [joined_user],
    }
    join_update = {'message': message}
    # Let the bot notice it's in the test group.
    conversation.multibot.dispatcher(conversation.bot, join_update)

    assert conversation.message('/admin modulestestbot moderator') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator: <b>Choose a group</b>
[-1001000001000 • My Group | /admin modulestestbot moderator -1001000001000]
[Back | /admin modulestestbot]
"""

    assert conversation.message('/admin modulestestbot moderator -1001000001000') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000: <b>Choose a field</b>
[calendars • Which calendars should be listed in /events? | /admin modulestestbot moderator -1001000001000 calendars]
[daily • Should I announce upcoming events once a day? | /admin modulestestbot moderator -1001000001000 daily]
[forward • Automatically forward messages from one chat to this one. | /admin modulestestbot moderator -1001000001000 forward]
[greeting • How should I greet people when they join? | /admin modulestestbot moderator -1001000001000 greeting]
[maxeventscount • How many events should be listed in /events? | /admin modulestestbot moderator -1001000001000 maxeventscount]
[maxeventsdays • How many days into the future should /events look? | /admin modulestestbot moderator -1001000001000 maxeventsdays]
[timezone • What time zone should be used in /events? | /admin modulestestbot moderator -1001000001000 timezone]
[Back | /admin modulestestbot moderator]
"""

    assert conversation.message('/admin modulestestbot moderator -1001000001000 bogus') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000: <b>Choose a field</b>

I can't set <code>bogus</code>.
[calendars • Which calendars should be listed in /events? | /admin modulestestbot moderator -1001000001000 calendars]
[daily • Should I announce upcoming events once a day? | /admin modulestestbot moderator -1001000001000 daily]
[forward • Automatically forward messages from one chat to this one. | /admin modulestestbot moderator -1001000001000 forward]
[greeting • How should I greet people when they join? | /admin modulestestbot moderator -1001000001000 greeting]
[maxeventscount • How many events should be listed in /events? | /admin modulestestbot moderator -1001000001000 maxeventscount]
[maxeventsdays • How many days into the future should /events look? | /admin modulestestbot moderator -1001000001000 maxeventsdays]
[timezone • What time zone should be used in /events? | /admin modulestestbot moderator -1001000001000 timezone]
[Back | /admin modulestestbot moderator]
"""

    assert conversation.message('/admin modulestestbot moderator -1001000001000 greeting') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › greeting: <b>Type a new value for greeting</b>

How should I greet people when they join?

Type your new value, or type "off" to disable/reset to default.
[Back | /admin modulestestbot moderator -1001000001000]
"""

    assert conversation.message(
        'Welcome to chat title, new users! <b>Initial</b> pinned message.') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000: <b>Choose a field</b>

Set <code>greeting</code> to <code>Welcome to chat title, new users! &lt;b&gt;Initial&lt;/b&gt; pinned message.</code>.
[calendars • Which calendars should be listed in /events? | /admin modulestestbot moderator -1001000001000 calendars]
[daily • Should I announce upcoming events once a day? | /admin modulestestbot moderator -1001000001000 daily]
[forward • Automatically forward messages from one chat to this one. | /admin modulestestbot moderator -1001000001000 forward]
[greeting (Welcome t…) • How should I greet people when they join? | /admin modulestestbot moderator -1001000001000 greeting]
[maxeventscount • How many events should be listed in /events? | /admin modulestestbot moderator -1001000001000 maxeventscount]
[maxeventsdays • How many days into the future should /events look? | /admin modulestestbot moderator -1001000001000 maxeventsdays]
[timezone • What time zone should be used in /events? | /admin modulestestbot moderator -1001000001000 timezone]
[Back | /admin modulestestbot moderator]
"""

    replies = conversation.raw_message('/dummy')
    assert replies == []
    conversation.multibot.dispatcher(conversation.bot, join_update)
    assert replies == []

    joined_user['is_bot'] = False

    replies = conversation.raw_message('/dummy')
    assert replies == []
    conversation.multibot.dispatcher(conversation.bot, join_update)
    assert replies == [
        {
            'chat_id': -1001000001000,
            'disable_notification': True,
            'disable_web_page_preview': True,
            'method': 'send_message',
            'parse_mode': 'HTML',
            'reply_to_message_id': 5000,
            'text': 'Welcome to My Group, <a href="tg://user?id=3000">User 3000</a>! <b>Initial</b> pinned message.',
        },
    ]  # yapf: disable

    message = {
        'from': adding_user,
        'chat': chat,
        'message_id': 5000,
        'pinned_message': {
            'message_id': 6000,
        },
    }
    pin_update = {'message': message}
    conversation.multibot.dispatcher(conversation.bot, pin_update)

    replies = conversation.raw_message('/dummy')
    assert replies == []
    conversation.multibot.dispatcher(conversation.bot, join_update)
    assert replies == [
        {
            'chat_id': -1001000001000,
            'disable_notification': True,
            'disable_web_page_preview': True,
            'method': 'send_message',
            'parse_mode': 'HTML',
            'reply_to_message_id': 5000,
            'text': 'Welcome to My Group, <a href="tg://user?id=3000">User 3000</a>! <b>Initial</b> <a href="https://t.me/c/1000001000/6000">pinned message</a>.',
        },
    ]  # yapf: disable

    # This group is now public!
    chat['username'] = 'mygroup'

    replies = conversation.raw_message('/dummy')
    assert replies == []
    conversation.multibot.dispatcher(conversation.bot, join_update)
    assert replies == [
        {
            'chat_id': -1001000001000,
            'disable_notification': True,
            'disable_web_page_preview': True,
            'method': 'send_message',
            'parse_mode': 'HTML',
            'reply_to_message_id': 5000,
            'text': 'Welcome to My Group, <a href="tg://user?id=3000">User 3000</a>! <b>Initial</b> <a href="https://t.me/mygroup/6000">pinned message</a>.',
        },
    ]  # yapf: disable

    assert conversation.message('/admin modulestestbot moderator -1001000001000 greeting') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › greeting: <b>Type a new value for greeting</b>

How should I greet people when they join?

<code>greeting</code> is currently <code>Welcome to chat title, new users! &lt;b&gt;Initial&lt;/b&gt; pinned message.</code>.

Type your new value, or type "off" to disable/reset to default.
[Back | /admin modulestestbot moderator -1001000001000]
"""

    assert conversation.message('Welcome! New message.') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000: <b>Choose a field</b>

Changed <code>greeting</code> from <code>Welcome to chat title, new users! &lt;b&gt;Initial&lt;/b&gt; pinned message.</code> to <code>Welcome! New message.</code>.
[calendars • Which calendars should be listed in /events? | /admin modulestestbot moderator -1001000001000 calendars]
[daily • Should I announce upcoming events once a day? | /admin modulestestbot moderator -1001000001000 daily]
[forward • Automatically forward messages from one chat to this one. | /admin modulestestbot moderator -1001000001000 forward]
[greeting (Welcome! …) • How should I greet people when they join? | /admin modulestestbot moderator -1001000001000 greeting]
[maxeventscount • How many events should be listed in /events? | /admin modulestestbot moderator -1001000001000 maxeventscount]
[maxeventsdays • How many days into the future should /events look? | /admin modulestestbot moderator -1001000001000 maxeventsdays]
[timezone • What time zone should be used in /events? | /admin modulestestbot moderator -1001000001000 timezone]
[Back | /admin modulestestbot moderator]
"""

    assert conversation.message('/admin modulestestbot moderator -1001000001000 greeting') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › greeting: <b>Type a new value for greeting</b>

How should I greet people when they join?

<code>greeting</code> is currently <code>Welcome! New message.</code>.

Type your new value, or type "off" to disable/reset to default.
[Back | /admin modulestestbot moderator -1001000001000]
"""

    assert conversation.message('OFF') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000: <b>Choose a field</b>

Unset <code>greeting</code> (was <code>Welcome! New message.</code>).
[calendars • Which calendars should be listed in /events? | /admin modulestestbot moderator -1001000001000 calendars]
[daily • Should I announce upcoming events once a day? | /admin modulestestbot moderator -1001000001000 daily]
[forward • Automatically forward messages from one chat to this one. | /admin modulestestbot moderator -1001000001000 forward]
[greeting • How should I greet people when they join? | /admin modulestestbot moderator -1001000001000 greeting]
[maxeventscount • How many events should be listed in /events? | /admin modulestestbot moderator -1001000001000 maxeventscount]
[maxeventsdays • How many days into the future should /events look? | /admin modulestestbot moderator -1001000001000 maxeventsdays]
[timezone • What time zone should be used in /events? | /admin modulestestbot moderator -1001000001000 timezone]
[Back | /admin modulestestbot moderator]
"""

    assert conversation.message(
        '/admin modulestestbot moderator -1001000001000 greeting OFF') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000: <b>Choose a field</b>

<code>greeting</code> is already unset.
[calendars • Which calendars should be listed in /events? | /admin modulestestbot moderator -1001000001000 calendars]
[daily • Should I announce upcoming events once a day? | /admin modulestestbot moderator -1001000001000 daily]
[forward • Automatically forward messages from one chat to this one. | /admin modulestestbot moderator -1001000001000 forward]
[greeting • How should I greet people when they join? | /admin modulestestbot moderator -1001000001000 greeting]
[maxeventscount • How many events should be listed in /events? | /admin modulestestbot moderator -1001000001000 maxeventscount]
[maxeventsdays • How many days into the future should /events look? | /admin modulestestbot moderator -1001000001000 maxeventsdays]
[timezone • What time zone should be used in /events? | /admin modulestestbot moderator -1001000001000 timezone]
[Back | /admin modulestestbot moderator]
"""



"""Tests for metabot.util.html."""

from metabot.util import html


def test_sanitize():
    """Test sanitization logic."""

    assert html.sanitize('') == ''
    assert html.sanitize('empty') == 'empty'
    assert html.sanitize('&#8592; &#x2190; &larr; \u2190') == '← ← ← ←'
    assert html.sanitize('&#38; &#x26; &amp;') == '&amp; &amp; &amp;'

    # From https://core.telegram.org/bots/api#formatting-options.
    text = """
<b>bold</b>, <strong>bold</strong>
<i>italic</i>, <em>italic</em>
<u>underline</u>, <ins>underline</ins>
<s>strikethrough</s>, <strike>strikethrough</strike>, <del>strikethrough</del>
<span class="tg-spoiler">spoiler</span>, <tg-spoiler>spoiler</tg-spoiler>
<b>bold <i>italic bold <s>italic bold strikethrough <span class="tg-spoiler">italic bold strikethrough spoiler</span></s> <u>underline italic bold</u></i> bold</b>
<a href="http://www.example.com/">inline URL</a>
<a href="tg://user?id=123456789">inline mention of a user</a>
<tg-emoji emoji-id="5368324170671202286">👍</tg-emoji>
<code>inline fixed-width code</code>
<pre>pre-formatted fixed-width code block</pre>
<pre><code class="language-python">pre-formatted fixed-width code block written in the Python programming language</code></pre>
<blockquote>Block quotation started\nBlock quotation continued\nThe last line of the block quotation</blockquote>
<blockquote expandable>Expandable block quotation started\nExpandable block quotation continued\nExpandable block quotation continued\nHidden by default part of the block quotation started\nExpandable block quotation continued\nThe last line of the block quotation</blockquote>
"""
    assert html.sanitize(text) == text
    assert html.sanitize('a<br>b<div>c</div><p>d</p>') == 'a\nb\nc\nd'
    assert html.sanitize('<code weird="true">text</code>') == '<code>text</code>'

    assert html.sanitize('<b>good</b> <bogus>bad</bogus>') == '<b>good</b> bad'
    assert html.sanitize('<b><bogus>nested</bogus></b>') == '<b>nested</b>'
    assert html.sanitize('<bogus><b>nested</b></bogus>') == '<b>nested</b>'
    assert html.sanitize('<b>broken <i>good</i>') == '<b>broken <i>good</i></b>'

    assert html.sanitize('&bogus;') == '&amp;bogus;'
    assert html.sanitize('&#bogus;') == '&amp;#bogus;'
    assert html.sanitize('&#xxbogus;') == '&amp;#xxbogus;'


def test_strip():
    """Test stripping logic."""

    assert html.sanitize('', True) == ''
    assert html.sanitize('empty', True) == 'empty'
    assert html.sanitize('&#8592; &#x2190; &larr; \u2190', True) == '← ← ← ←'
    assert html.sanitize('&#38; &#x26; &amp;', True) == '&amp; &amp; &amp;'

    # From https://core.telegram.org/bots/api#formatting-options.
    text = """
<b>bold</b>, <strong>bold</strong>
<i>italic</i>, <em>italic</em>
<u>underline</u>, <ins>underline</ins>
<s>strikethrough</s>, <strike>strikethrough</strike>, <del>strikethrough</del>
<span class="tg-spoiler">spoiler</span>, <tg-spoiler>spoiler</tg-spoiler>
<b>bold <i>italic bold <s>italic bold strikethrough <span class="tg-spoiler">italic bold strikethrough spoiler</span></s> <u>underline italic bold</u></i> bold</b>
<a href="http://www.example.com/">inline URL</a>
<a href="tg://user?id=123456789">inline mention of a user</a>
<tg-emoji emoji-id="5368324170671202286">👍</tg-emoji>
<code>inline fixed-width code</code>
<pre>pre-formatted fixed-width code block</pre>
<pre><code class="language-python">pre-formatted fixed-width code block written in the Python programming language</code></pre>
<blockquote>Block quotation started\nBlock quotation continued\nThe last line of the block quotation</blockquote>
<blockquote expandable>Expandable block quotation started\nExpandable block quotation continued\nExpandable block quotation continued\nHidden by default part of the block quotation started\nExpandable block quotation continued\nThe last line of the block quotation</blockquote>
"""
    assert html.sanitize(text, True) == """
bold, bold
italic, italic
underline, underline
strikethrough, strikethrough, strikethrough
spoiler, spoiler
bold italic bold italic bold strikethrough italic bold strikethrough spoiler underline italic bold bold
inline URL
inline mention of a user
👍
inline fixed-width code
pre-formatted fixed-width code block
pre-formatted fixed-width code block written in the Python programming language
Block quotation started\nBlock quotation continued\nThe last line of the block quotation
Expandable block quotation started\nExpandable block quotation continued\nExpandable block quotation continued\nHidden by default part of the block quotation started\nExpandable block quotation continued\nThe last line of the block quotation
"""
    assert html.sanitize('first<br>second', True) == 'first\nsecond'
    assert html.sanitize('<a weird="true">text</a>', True) == 'text'

    assert html.sanitize('<b>good</b> <span>bad</span>', True) == 'good bad'
    assert html.sanitize('<b><span>nested</span></b>', True) == 'nested'
    assert html.sanitize('<span><b>nested</b></span>', True) == 'nested'
    assert html.sanitize('<b>broken <i>good</i>', True) == 'broken good'

    assert html.sanitize('&bogus;', True) == '&amp;bogus;'
    assert html.sanitize('&#bogus;', True) == '&amp;#bogus;'
    assert html.sanitize('&#xxbogus;', True) == '&amp;#xxbogus;'


def test_truncate():
    """Test truncating logic."""

    assert html.truncate('', 10) == ''
    assert html.truncate('123456789012345', 10) == '1234567890'
    assert html.truncate('<b>123456789012345</b>', 10) == '<b>1234567890</b>'
    assert html.truncate('123<b>456<i>789012345', 10) == '123<b>456<i>7890</i></b>'
    assert html.truncate('&lt;' * 100, 10) == '&lt;' * 10
    assert html.truncate('&#x2022;' * 100, 10) == '\u2022' * 10

    # I originally had a short circuit in _HTMLSanitizer.__append:
    #   if len(data) <= self.__remaining:
    #       self.__remaining -= len(data)
    #       return self.__pieces.append(data)
    # which would overcount the lengths of entities if they were part of a fragment that otherwise
    # didn't trip the length limit. So this would work because the fragment tripped the limit:
    assert html.truncate('&lt;2345678901234567890', 10) == '&lt;234567890'
    # but this would truncate to '&lt;2345<b>67</b>' because it counted '&lt;' as 4 instead of 1
    # while processing '&lt;2345':
    assert html.truncate('&lt;2345<b>678901234567890', 10) == '&lt;2345<b>67890</b>'


def test_MessageEntity_cpp_quirks():  # pylint: disable=invalid-name
    """Verify sync with https://github.com/tdlib/td/blob/master/td/telegram/MessageEntity.cpp."""

    assert html.sanitize('<a foo=1 href=2 bar=3>b</a>') == '<a href="2">b</a>'
    # Technically, '<a>https://example.com/</a>' would be considered valid, but I don't want to
    # have to validate URLs.
    assert html.sanitize('<a foo=1 bar=3>b</a>') == 'b'

    assert html.sanitize('<SPAN CLASS="tg-spoiler">x</span>') == '<span class="tg-spoiler">x</span>'
    assert html.sanitize('<span class="other">x</span>') == 'x'
    assert html.sanitize('<span>x</span>') == 'x'

    assert html.sanitize('<blockquote>quote</blockquote>') == '<blockquote>quote</blockquote>'
    assert html.sanitize(
        '<blockquote other=1>quote</blockquote>') == '<blockquote>quote</blockquote>'
    assert html.sanitize(
        '<blockquote expandable>quote</blockquote>') == '<blockquote expandable>quote</blockquote>'
    assert html.sanitize('<blockquote expandable=true>quote</blockquote>'
                        ) == '<blockquote expandable="true">quote</blockquote>'

    assert html.sanitize('<tg-emoji>emoji</tg-emoji>') == 'emoji'
    assert html.sanitize(
        '<tg-emoji emoji-id=1>emoji</tg-emoji>') == '<tg-emoji emoji-id="1">emoji</tg-emoji>'
    assert html.sanitize('<tg-emoji emoji-id=0>emoji</tg-emoji>') == 'emoji'
    assert html.sanitize('<tg-emoji emoji-id=10000000000000000000>emoji</tg-emoji>') == 'emoji'
    assert html.sanitize('<tg-emoji emoji-id=dummy>emoji</tg-emoji>') == 'emoji'

    assert html.sanitize('<code>code</code>') == '<code>code</code>'
    assert html.sanitize(
        '<code class="language-python">code</code>') == '<code class="language-python">code</code>'
    assert html.sanitize('<code class="other">code</code>') == '<code>code</code>'


def test_wellformed_close():
    """Verify well-formed but Telegram-incompatible markup does not close other tags."""

    assert html.sanitize(
        'hi <code>there <span>friend</span> still') == 'hi <code>there friend still</code>'



"""Return the list of commands and other bot features."""

import collections

ALIASES = ('command', 'commands', 'help', 'start')


def moddispatch(ctx, msg, modconf):  # pylint: disable=missing-docstring
    if ctx.type in ('message', 'callback_query') and ctx.command in ALIASES:
        return default(ctx, msg, modconf)

    return False


def default(ctx, msg, modconf):  # pylint: disable=missing-docstring
    ctx.private = True
    sections = collections.defaultdict(set)

    hidden = set(modconf.get('hidden', '').split())

    for modname, module in ctx.bot.multibot.modules.items():
        if modname in hidden:
            continue
        modhelp = getattr(module, 'modhelp', None)
        if modhelp:
            modhelp(ctx, ctx.bot.config['issue37'][modname], sections)

    if not sections:
        msg.add("I don't have much documentation\u2014check with a bot admin!")
    else:
        for name, lines in sorted(sections.items()):
            msg.add('<b>%s</b>', name.title())
            for line in sorted(lines):
                msg.add('%s', line)


def admin(frame):  # pylint: disable=too-many-branches
    """Handle /admin BOTNAME help."""

    ctx, msg, modconf = frame.ctx, frame.msg, frame.value
    action, _, modname = frame.text.partition(' ')

    hidden = set(modconf.get('hidden', '').split())

    if action == 'hide':
        if modname in hidden:
            msg.add('<code>%s</code> is already hidden!', modname)
        else:
            msg.add('<code>%s</code> is now hidden.', modname)
            hidden.add(modname)
    elif action == 'unhide':
        if modname not in hidden:
            msg.add('<code>%s</code> is not hidden!', modname)
        else:
            msg.add('<code>%s</code> is now visible.', modname)
            hidden.remove(modname)

    if hidden:
        modconf['hidden'] = ' '.join(sorted(hidden))
    else:
        modconf.pop('hidden')

    modules = collections.defaultdict(lambda: collections.defaultdict(set))
    for modname, module in ctx.bot.multibot.modules.items():
        modhelp = getattr(module, 'modhelp', None)
        if modhelp:
            modhelp(ctx, ctx.bot.config['issue37'][modname], modules[modname])

    msg.action = 'Select a module'
    for modname, sections in sorted(modules.items()):
        label = '%s (%s)' % (modname, ' \u2022 '.join(sorted(sections['commands'])))
        if modname in hidden:
            msg.button('Show ' + label, 'unhide ' + modname)
        elif sections:
            msg.button('Hide ' + label, 'hide ' + modname)



"""Quick routines to convert machine-friendly data types to human-friendly strings."""

import calendar
import datetime
import time as _time

from metabot.util import unicodeutil


def _now_copy_tz(dtime):
    now = datetime.datetime.fromtimestamp(_time.time(), getattr(dtime, 'tzinfo', None))
    return isinstance(dtime, datetime.datetime) and now or now.date()


def date(dtime, base=None):  # pylint: disable=too-many-branches
    """Convert the given datetime.date/datetime.datetime into a human-friendly string."""

    if not base:
        base = _now_copy_tz(dtime)  # pragma: no cover
    day = dtime.strftime('%a')
    mon = dtime.strftime('%b')
    dom = dayofmonth(dtime.day)
    if dtime.year != base.year:
        text = '%s, %s (%i) %s' % (day, mon, dtime.year, dom)
    elif dtime.month != base.month:
        text = '%s, %s %s' % (day, mon, dom)
    else:
        text = '%s %s' % (day, dom)
    if isinstance(dtime, datetime.datetime):
        text = '%s, %s' % (text, time(dtime))
    return text


def dayofmonth(dom):
    """Return a day of the month in ordinal form (1st, 11th, etc.)."""

    ending = 'th'
    if not 11 <= dom <= 13:
        ones = dom % 10
        if ones == 1:
            ending = 'st'
        elif ones == 2:
            ending = 'nd'
        elif ones == 3:
            ending = 'rd'
    return '%s%s' % (dom, unicodeutil.superscript(ending))


def _nextmonth(dt, months=1):
    day = dt.day

    while months > 0:
        months -= 1
        if dt.month == 12:
            year = dt.year + 1
            month = 1
        else:
            year = dt.year
            month = dt.month + 1
        _, ndays = calendar.monthrange(year, month)
        dt = dt.replace(year=year, month=month, day=min(day, ndays))
    while months < 0:
        months += 1
        if dt.month == 1:
            year = dt.year - 1
            month = 12
        else:
            year = dt.year
            month = dt.month - 1
        _, ndays = calendar.monthrange(year, month)
        dt = dt.replace(year=year, month=month, day=min(day, ndays))
    return dt


def _nextyear(dt, years=1):
    year = dt.year + years
    _, ndays = calendar.monthrange(year, dt.month)
    return dt.replace(year=year, day=min(dt.day, ndays))


def _howrecent(start, end, base=None):  # pylint: disable=too-many-branches,too-many-return-statements,too-many-statements
    if not base:
        base = _now_copy_tz(start)  # pragma: no cover
    if start <= base <= end:
        return '\u2b50 NOW'
    if isinstance(base, datetime.datetime):
        if base < start:
            seconds = int((start - base).total_seconds())
            if seconds <= 60 * 60 * 24:
                hours, seconds = divmod(seconds, 3600)
                minutes, seconds = divmod(seconds, 60)
                if not hours:
                    prefix = f'{minutes} min'
                elif minutes:
                    prefix = f'{hours}h{minutes}m'
                else:
                    prefix = plural(hours, 'hour')
                return f'\U0001f51c {prefix}'

        start = start.date()
        base = base.date()

    delta = start - base
    if delta.days == 0:
        return 'today'
    if delta.days == 1:
        return 'tomorrow'
    if delta.days == -1:
        return 'yesterday'

    if start < base:
        left = start
        right = base
    else:
        left = base
        right = start

    years = 0
    left = _nextyear(left)
    while left <= right:
        years += 1
        left = _nextyear(left)
    left = _nextyear(left, -1)

    months = 0
    left = _nextmonth(left)
    while left <= right:
        months += 1
        left = _nextmonth(left)
    left = _nextmonth(left, -1)

    weeks, days = divmod((right - left).days, 7)

    if years:
        prefix = f'{years}y'
        if months:
            prefix = f'{prefix}{months}m'
    elif months:
        prefix = f'{months}m'
        if weeks:
            prefix = f'{prefix}{weeks}w'
    elif weeks:
        prefix = f'{weeks}w'
        if days:
            prefix = f'{prefix}{days}d'
    else:
        prefix = f'{days} days'

    if delta.days < 0:
        prefix += ' ago'
    return prefix


def howrecent(start, end, base=None):
    """Convert the delta between start and base into a human-friendly string."""

    return unicodeutil.superscript(_howrecent(start, end, base=base))


def list(arr):  # pylint: disable=redefined-builtin
    """Convert the given list into a human-friendly string."""

    if len(arr) == 1:
        return arr[0]
    if len(arr) == 2:
        return '%s and %s' % (arr[0], arr[1])
    return '%s, and %s' % (', '.join(arr[:-1]), arr[-1])


def plural(num, noun, fmtstr='%s %s'):
    """Return '1 noun', '2 nouns', etc."""

    if num == 1:
        return fmtstr % (num, noun)
    return fmtstr % (num, noun + 's')


def range(start, end, base=None):  # pylint: disable=redefined-builtin
    """Convert the given datetime.date/datetime.datetime objects into a human-friendly string."""

    if not base:
        base = _now_copy_tz(start)
    text = date(start, base=base)
    if start == end:
        return text
    if isinstance(start, datetime.datetime):
        if (start.year != end.year or start.month != end.month or
                end - start >= datetime.timedelta(days=1)):
            return '%s \u2013 %s' % (text, date(end, base=start.date()))
        if start.day != end.day or (start.hour < 12) != (end.hour < 12):
            return '%s \u2013 %s' % (text, time(end))
        return '%s\u2013%s' % (text[:-2], time(end))
    if start.year != end.year or start.month != end.month:
        return '%s \u2013 %s' % (text, date(end, base=start))
    return '%s\u2013%s' % (text[:-2], dayofmonth(end.day))


def time(dtime):
    """Convert the given datetime.datetime/datetime.time into a human-friendly string."""

    text = '%i' % (((dtime.hour - 1) % 12) + 1)
    if dtime.minute:
        text = '%s:%02i' % (text, dtime.minute)
    return text + unicodeutil.superscript(dtime.strftime('%P'))



"""Tests for metabot.modules.admin."""

import pytest

from metabot.modules import admin


@pytest.fixture
def conversation(build_conversation):  # pylint: disable=missing-docstring
    return build_conversation(admin)


# pylint: disable=line-too-long


def test_invalid_user(conversation):  # pylint: disable=redefined-outer-name
    """Verify the bot doesn't respond to the /admin command from non-admins."""

    error_message = """\
[chat_id=2000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin

Hi! You aren't one of my admins. If you should be, ask a current admin to add you by opening a chat with me (@modulestestbot) and typing:

<pre>/admin modulestestbot admin add 2000</pre>
"""

    assert conversation.message('/admin', user_id=2000) == error_message
    assert conversation.message('/admin dummy', user_id=2000) == error_message
    assert conversation.message('/admin modulestestbot admin add 2000',
                                user_id=2000) == error_message


def test_default(conversation):  # pylint: disable=redefined-outer-name
    """Verify the /admin command gives its lovely menu of subcommands."""

    assert conversation.message('/notadmin') == ''

    assert conversation.message('/admin') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin: <b>Choose a bot</b>

This is a metabot! Check out https://github.com/nmlorg/metabot/issues to keep track of bugs and features.

To configure your bot, choose its username:
[modulestestbot | /admin modulestestbot]
"""

    assert conversation.message('/admin modulestestbot') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot: <b>Choose a module</b>
[admin • Manage the admin list | /admin modulestestbot admin]
[help • Return the list of commands and other bot features | /admin modulestestbot help]
[Back | /admin]
"""


def test_whoami(conversation):  # pylint: disable=redefined-outer-name
    """Test /whoami."""

    assert conversation.message('/whoami') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
<code>1000</code>
"""


def test_bootstrap(conversation):  # pylint: disable=redefined-outer-name
    """Test /_bootstrap."""

    assert conversation.bot.config['issue37'].pop('admin') == {'admins': [1000]}
    assert len(admin.BOOTSTRAP_TOKEN) == 32
    admin.BOOTSTRAP_TOKEN = 'bootstraptest'

    assert conversation.message('/_bootstrap') == ''
    assert conversation.bot.config['issue37'].get('admin') is None

    assert conversation.message('/_bootstrap bogus') == ''
    assert conversation.bot.config['issue37'].get('admin') is None

    assert conversation.message('/_bootstrap bootstraptest') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Added 1000 to the admin list.
"""
    assert conversation.bot.config['issue37'].get('admin') == {'admins': [1000]}

    assert conversation.message('/_bootstrap bootstraptest') == ''
    assert conversation.bot.config['issue37'].get('admin') == {'admins': [1000]}

    assert conversation.message('/_bootstrap bootstraptest', user_id=2000) == ''
    assert conversation.bot.config['issue37'].get('admin') == {'admins': [1000]}


def test_admins(conversation):  # pylint: disable=redefined-outer-name
    """Verify /admin's own configurator."""

    assert conversation.message('/admin modulestestbot admin') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › admin: <b>Choose an admin</b>

Forward a message from a user to add or remove them, or select an existing admin to remove.
[Back | /admin modulestestbot]
"""

    assert conversation.message('bogus value') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › admin: <b>Choose an admin</b>

I'm not sure what <code>bogus value</code> is—it's not a user id!

Forward a message from a user to add or remove them, or select an existing admin to remove.
[Back | /admin modulestestbot]
"""

    assert conversation.message('1000') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › admin: <b>Choose an admin</b>

You can't remove yourself from the admin list.

Forward a message from a user to add or remove them, or select an existing admin to remove.
[Back | /admin modulestestbot]
"""

    assert conversation.message('2000') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › admin: <b>Choose an admin</b>

Added 2000 to the admin list.

Forward a message from a user to add or remove them, or select an existing admin to remove.
[Remove 2000 | /admin modulestestbot admin 2000]
[Back | /admin modulestestbot]
"""

    assert conversation.message('/dummy', user_id=2000) == ''

    assert conversation.message('/admin modulestestbot admin bogus value') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › admin: <b>Choose an admin</b>

I'm not sure what <code>bogus value</code> is—it's not a user id!

Forward a message from a user to add or remove them, or select an existing admin to remove.
[Remove User2000 (2000) | /admin modulestestbot admin 2000]
[Back | /admin modulestestbot]
"""

    assert conversation.message('/admin modulestestbot admin 2000') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › admin: <b>Choose an admin</b>

Removed 2000 from the admin list.

Forward a message from a user to add or remove them, or select an existing admin to remove.
[Back | /admin modulestestbot]
"""

    assert conversation.message('/whoami', forward_date=1000) == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › admin: <b>Choose an admin</b>

I can't tell exactly who sent that message, possibly because the sender enabled <a href="https://telegram.org/blog/unsend-privacy-emoji#anonymous-forwarding">Anonymous Forwarding</a>. You can either ask them to temporarily re-enable normal forwarding, or ask them to open a chat with me, type <code>/whoami</code>, then forward or copy/paste the number they receive to you (to copy/paste on to me).

Forward a message from a user to add or remove them, or select an existing admin to remove.
[Back | /admin modulestestbot]
"""

    assert conversation.message('/whoami', forward_date=1000, forward_from=2000) == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › admin: <b>Choose an admin</b>

Added 2000 to the admin list.

Forward a message from a user to add or remove them, or select an existing admin to remove.
[Remove User2000 (2000) | /admin modulestestbot admin 2000]
[Back | /admin modulestestbot]
"""



"""Create custom commands that just return fixed messages."""

from metabot.util import adminui


def modhelp(unused_ctx, modconf, sections):  # pylint: disable=missing-docstring
    for command, data in modconf.items():
        if not data.get('text'):
            continue
        message = data['text'].replace('\n', ' ')
        if len(message) > 30:
            message = message[:29] + '\u2026'
        sections['commands'].add('/%s \u2013 "%s"' % (command, message))


def moddispatch(ctx, msg, modconf):  # pylint: disable=missing-docstring
    if (ctx.type in ('message', 'callback_query') and ctx.command in modconf and
            modconf[ctx.command].get('text')):
        return echo(ctx, msg, modconf[ctx.command])

    return False


def echo(ctx, msg, data):  # pylint: disable=missing-docstring
    ctx.private = data.get('private')
    if not data.get('paginate'):
        msg.add(data['text'])
    else:
        lines = [line for line in data['text'].splitlines() if line.strip()]
        page = ctx.text.isdigit() and int(ctx.text) or 1
        for line in lines[:page]:
            msg.add('%s', line)
        if page < len(lines):
            msg.button('More (%i/%i)' % (page, len(lines)), '/%s %i' % (ctx.command, page + 1))


def admin(frame):
    """Handle /admin BOTNAME echo."""

    msg = frame.msg
    menu = adminui.Menu()
    for command, data in frame.value.items():
        menu.add(command, desc=data.get('text', '').replace('\n', ' '))
    menu.add(None)
    frame, handler = menu.select(frame)
    if not handler:
        msg.action = 'Choose a command'
        msg.add(
            "Type the name of a command to add (like <code>rules</code>\u2014don't include a slash "
            'at the beginning!), or select an existing echo.')
        return menu.display(frame, what='command')

    msg.path(frame.field)

    adminui.Menu(
        ('text', adminui.freeform,
         'The message, sticker, or image to send in response to /%s.' % frame.field),
        ('paginate', adminui.bool, 'For multiline messages, display just one line at a time?'),
        ('private', adminui.bool, 'Send the message in group chats, or just in private?'),
    ).handle(frame)



"""Return a Calendar object for the given calid, constructing it if necessary."""

import importlib
import logging

from metabot.util import pickleutil

CALENDAR_TYPES = {}
CALENDARS = {}


def get(calid):
    """Return a Calendar object for the given calid, constructing it if necessary."""

    cal = CALENDARS.get(calid)
    if not cal:
        proto = calid.split(':', 1)[0]
        calendar_type = CALENDAR_TYPES.get(proto)
        if not calendar_type:
            module = importlib.import_module('metabot.calendars.' + proto)

            # pylint: disable=used-before-assignment
            class _CachingCalendar(_CachingCalendarMixin, module.Calendar):
                pass

            CALENDAR_TYPES[proto] = calendar_type = _CachingCalendar
        CALENDARS[calid] = cal = calendar_type(calid)
    return cal


class _CachingCalendarMixin:
    _cache_dir = 'calendars'

    def __init__(self, calid):
        super().__init__(calid)
        self.__fname = '%s/%s.pickle' % (self._cache_dir, self.calcode)
        data = pickleutil.load(self.__fname)
        if data:
            self.__dict__.update(data)

    def __save(self):
        logging.info('Rewriting %r.', self.__fname)
        self.__dict__.update(pickleutil.dump(self.__fname, self.__dict__))

    def poll(self):  # pylint: disable=missing-docstring
        if super().poll():
            self.__save()
            return True

    def add(self, local):  # pylint: disable=missing-docstring
        ret = super().add(local)
        self.__save()
        return ret

    def remove(self, local_id):  # pylint: disable=missing-docstring
        ret = super().remove(local_id)
        self.__save()
        return ret

    def update(self, local_id, local):  # pylint: disable=missing-docstring
        ret = super().update(local_id, local)
        self.__save()
        return ret



"""Tests for metabot.botconf."""

import pytest
import yaml

from metabot import botconf


# See https://github.com/yaml/pyyaml/issues/202.
@pytest.mark.filterwarnings('ignore:Using or importing the ABCs from')
def test_save_load(tmpdir):
    """Verify MultiBot can start with no config, can have a bot added, and can restart."""

    conffile = tmpdir.join('autogen.yaml')

    conf = botconf.BotConf(confdir=tmpdir.strpath)
    conf['autogen']['alpha']['bravo'] = {'charlie': 'delta'}
    conf['autogen']['alpha']['echo'] = [2, 4, 6]
    conf.save()

    assert yaml.safe_load(conffile.read()) == {
        'alpha': {
            'bravo': {
                'charlie': 'delta',
            },
            'echo': [2, 4, 6],
        },
    }



"""Quick utility to load all non-test-related modules in a package."""

import importlib
import pkgutil


def load_modules(package):
    """Load all non-test-related modules in a package."""

    if isinstance(package, str):
        package = importlib.import_module(package)
    modules = set()
    for _, name, _ in pkgutil.iter_modules(package.__path__):
        if name != 'conftest' and not name.startswith('test_'):
            modules.add(importlib.import_module('%s.%s' % (package.__name__, name)))
    return modules



"""Announce upcoming events once a day."""

import collections
import datetime
import logging
import operator
import random
import re
import time

import ntelebot

from metabot.util import eventutil
from metabot.util import html
from metabot.util import humanize
from metabot.util import pickleutil

PERIOD = 60 * 10  # Run modinit.periodic every 10 minutes.


def modinit(multibot):  # pylint: disable=missing-docstring

    def queue():
        now = time.time()
        nextrun = (now // PERIOD + 1) * PERIOD
        jitter = random.random() * 5
        multibot.loop.queue.putwhen(nextrun + jitter, periodic)

    if multibot.conf.confdir:
        recordsfname = multibot.conf.confdir + '/daily.pickle'
        records = pickleutil.load(recordsfname) or {}

        for key, record in records.items():
            if len(record) == 3:
                records[key] += ('', '')
    else:
        recordsfname = None
        records = {}

    def periodic():
        logging.info('Running periodic.')
        try:
            multibot.multical.poll()
            _daily_messages(multibot, records)
            if recordsfname:
                pickleutil.dump(recordsfname, records)
        finally:
            queue()

    queue()


class AnnouncementConf:  # pylint: disable=too-few-public-methods
    """A group's announcement configuration."""

    def __init__(self, calconf, dailyconf):
        self.calconf = calconf
        self.dow = dailyconf.get('dow', 0)
        self.hour = dailyconf.get('hour')
        self.pin = dailyconf.get('pin', 0)
        self.preambles = dailyconf.get('text', '').splitlines()

    def get_events(self, bot, eventtime, base, *, countdown=True):
        """Get (and format) events for the given time."""

        calconf = self.calconf
        events = calconf.get_events(bot, when=eventtime)
        if self.preambles:
            preamble = self.preambles[int(eventtime / (60 * 60 * 24)) % len(self.preambles)]
        else:
            preamble = ''
        text = _generate_preamble(preamble, events)
        if events:
            ev = eventutil.format_events(bot,
                                         events,
                                         calconf.tzinfo,
                                         base=base,
                                         countdown=countdown)
            text = f'{text}\n\n{ev}'
        return events, text


class Announcement(collections.namedtuple('Announcement', 'time events message text suffix')):
    """The most recent announcement."""


def _daily_messages(multibot, records):  # pylint: disable=too-many-branches,too-many-locals,too-many-statements
    now = time.time()
    # If running at 11:22:33.444, act as if we're running at exactly 11:20:00.000.
    period = int(now // PERIOD * PERIOD)
    startofhour = now // 3600

    for botuser, botconf in multibot.conf['bots'].items():  # pylint: disable=too-many-nested-blocks
        for groupid, groupconf in botconf['issue37']['moderator'].items():
            calconf = eventutil.CalendarConf(groupconf)
            annconf = AnnouncementConf(calconf, groupconf['daily'])
            if not calconf.tzinfo or not isinstance(annconf.hour, int):
                continue

            nowdt = datetime.datetime.fromtimestamp(now, calconf.tzinfo)
            perioddt = datetime.datetime.fromtimestamp(period, calconf.tzinfo)
            key = (botuser, groupid)
            if key in records:
                last = Announcement(*records[key])
            else:
                last = None

            # See https://github.com/nmlorg/metabot/issues/26.
            bot = ntelebot.bot.Bot(botconf['issue37']['telegram']['token'])
            bot.multibot = multibot
            bot._username = botuser  # pylint: disable=protected-access

            if perioddt.hour == annconf.hour and not annconf.dow & 1 << perioddt.weekday() and (
                    not last or startofhour > last.time // 3600):
                events, text = annconf.get_events(bot, period, perioddt)
                if events:
                    url = eventutil.get_image(events[0], botconf)
                    message = reminder_send(bot, groupid, text, url)
                    if message:
                        if annconf.pin:
                            try:
                                bot.pin_chat_message(chat_id=message['chat']['id'],
                                                     message_id=message['message_id'],
                                                     disable_notification=True)
                                message['pinned'] = True
                            except ntelebot.errors.Error:
                                logging.exception('While pinning %s/%s:', message['chat']['id'],
                                                  message['message_id'])

                        records[key] = (period, [event.copy() for event in events], message, text,
                                        '')

                        if last:
                            text = annconf.get_events(bot, last.time, perioddt, countdown=False)[1]
                            reminder_edit(bot, last.message, text)

                            if last.message.get('pinned'):
                                try:
                                    bot.unpin_chat_message(chat_id=last.message['chat']['id'],
                                                           message_id=last.message['message_id'])
                                except ntelebot.errors.Error:
                                    logging.exception('While unpinning %s/%s:',
                                                      last.message['chat']['id'],
                                                      last.message['message_id'])
                    continue

            if last:
                events, text = annconf.get_events(bot, last.time, perioddt)
                edits = diff_events(multibot, calconf.tzinfo, perioddt, last.events, events)

                suffix = last.suffix
                if edits:
                    updtext = 'Updated:\n' + '\n'.join(edits)
                    try:
                        updmessage = bot.send_message(
                            chat_id=groupid,
                            reply_to_message_id=last.message['message_id'],
                            text=_truncate(updtext, ntelebot.limits.message_text_length_max),
                            parse_mode='HTML',
                            disable_web_page_preview=True,
                            disable_notification=True)
                    except ntelebot.errors.Error:
                        logging.exception('While sending to %s:\n%s', groupid, updtext)
                    else:
                        suffix = 'Updated ' + humanize.time(nowdt)
                        groupidnum = int(groupid)
                        if -1002147483647 <= groupidnum < -1000000000000:
                            suffix = '<a href="https://t.me/c/%s/%s">%s</a>' % (
                                -1000000000000 - groupidnum, updmessage['message_id'], suffix)

                if text != last.text or suffix != last.suffix:
                    newtext = text
                    if suffix:
                        newtext = f'{newtext}\n\n[{suffix}]'
                    message = reminder_edit(bot, last.message, newtext)
                    if message:
                        records[key] = (last.time, [event.copy() for event in events], message,
                                        text, suffix)


def reminder_send(bot, groupid, text, photo):
    """Send a message with the given photo + caption, falling back to plain text."""

    base = {
        'chat_id': groupid,
        'disable_notification': True,
        'parse_mode': 'HTML',
    }

    logging.info('Sending reminder to %s.', groupid)
    try:
        if photo:
            try:
                return bot.send_photo(**base, photo=photo, caption=text)
            except ntelebot.errors.TooLong:  # See https://github.com/nmlorg/metabot/issues/76.
                logging.info('Downgrading to plain text.')

        return bot.send_message(**base,
                                text=_truncate(text, ntelebot.limits.message_text_length_max),
                                disable_web_page_preview=True)
    except ntelebot.errors.Error:
        logging.exception('While sending to %s:\n%s', groupid, text)


def reminder_edit(bot, lastmessage, text):
    """Edit a photo caption/plain message."""

    groupid = lastmessage['chat']['id']
    message_id = lastmessage['message_id']
    base = {
        'chat_id': groupid,
        'message_id': message_id,
        'parse_mode': 'HTML',
    }

    logging.info('Editing reminder %s/%s.', groupid, message_id)
    try:
        if lastmessage.get('caption'):
            truncated = _truncate(text, ntelebot.limits.message_caption_length_max)
            message = bot.edit_message_caption(**base, caption=truncated)
        else:
            truncated = _truncate(text, ntelebot.limits.message_text_length_max)
            message = bot.edit_message_text(**base, text=truncated, disable_web_page_preview=True)
        if lastmessage.get('pinned'):
            message['pinned'] = True
        return message
    except ntelebot.errors.Unmodified:
        logging.exception('While editing %s/%s:\n%s', groupid, message_id, text)
        return lastmessage  # See https://github.com/nmlorg/metabot/issues/108.
    except ntelebot.errors.Error:
        logging.exception('While editing %s/%s:\n%s', groupid, message_id, text)


def _truncate(rawtext, length):
    text = html.truncate(rawtext, length)
    if text != rawtext:
        logging.info('Truncated:\n- %r\n+ %r', rawtext, text)
    return text


def diff_events(multibot, tzinfo, base, lastevents, events):  # pylint: disable=too-many-locals
    """Return a list of differences between lastevents and events."""

    lastmap = {event['local_id']: event for event in lastevents}
    curmap = {event['local_id']: event for event in events}
    bothevents = events.copy()
    bothevents.extend(event for event in lastevents if event['local_id'] not in curmap)
    bothevents.sort(key=operator.itemgetter('start', 'end', 'summary', 'local_id'))
    edits = []
    for event in bothevents:
        title = html.escape(event['summary'])

        lastevent = lastmap.get(event['local_id'])
        if not lastevent:
            edits.append(f'\u2022 <b>{title}</b> was added.')
            continue

        # See https://github.com/nmlorg/metabot/issues/75.
        event = multibot.multical.get_event(event['local_id'])[1]
        if not event:
            edits.append(f'\u2022 <s>{title}</s> was removed.')
            continue

        mentioned_event = False
        pieces = []

        diff = _quick_diff(html.escape(lastevent['summary']), html.escape(event['summary']))
        if diff:
            pieces.append(f'<s>{diff[0]}</s> is now called <b>{diff[1]}</b>')
            mentioned_event = True

        diff = _diff_time(lastevent['start'], lastevent['end'], event['start'], event['end'],
                          tzinfo, base)
        if diff:
            pieces.append(diff)

        diff = _quick_diff(html.escape(lastevent['location'].split(',', 1)[0]),
                           html.escape(event['location'].split(',', 1)[0]))
        if diff:
            pieces.append(f'was moved from <s>{diff[0]}</s> to <b>{diff[1]}</b>')

        diff = _quick_diff(html.sanitize(lastevent['description'], strip=True),
                           html.sanitize(event['description'], strip=True))
        if diff:
            if pieces:
                prefix = 'its description'
            else:
                prefix = f'The description of <b>{title}</b>'
                mentioned_event = True
            pieces.append(f'{prefix} was changed from <s>{diff[0]}</s> to <b>{diff[1]}</b>')

        if pieces:
            text = humanize.list(pieces)
            if not mentioned_event:
                text = f'<b>{title}</b> {text}'
            edits.append(f'\u2022 {text}.')

    return edits


def _diff_time(laststart, lastend, curstart, curend, tzinfo, base):  # pylint: disable=too-many-arguments,too-many-locals,too-many-positional-arguments,too-many-return-statements
    laststartdt = datetime.datetime.fromtimestamp(laststart, tzinfo)
    lastenddt = datetime.datetime.fromtimestamp(lastend, tzinfo)
    curstartdt = datetime.datetime.fromtimestamp(curstart, tzinfo)
    curenddt = datetime.datetime.fromtimestamp(curend, tzinfo)

    laststartstr = humanize.date(laststartdt, base=base)
    lastendstr = humanize.date(lastenddt, base=base)
    curstartstr = humanize.date(curstartdt, base=base)
    curendstr = humanize.date(curenddt, base=base)

    start_earlier = curstart < laststart
    start_later = curstart > laststart
    start_same = curstart == laststart
    end_earlier = curend < lastend
    end_later = curend > lastend
    end_same = curend == lastend
    duration_same = lastend - laststart == curend - curstart

    # pylint: disable=line-too-long
    if not start_same and not end_same and not duration_same:  # Everything changed.
        return f'now starts at <b>{curstartstr}</b> and ends at <b>{curendstr}</b> (was <s>{laststartstr} to {lastendstr}</s>)'
    if start_earlier:
        if duration_same:
            return f'was moved up to <b>{curstartstr}</b> (from <s>{laststartstr}</s>; same duration)'
        return f'is starting earlier at <b>{curstartstr}</b> (instead of <s>{laststartstr}</s>; same end)'
    if start_later:
        if duration_same:
            return f'was moved back to <b>{curstartstr}</b> (from <s>{laststartstr}</s>; same duration)'
        return f'is starting later at <b>{curstartstr}</b> (instead of <s>{laststartstr}</s>; same end)'
    if end_earlier:
        return f'was shortened to <b>{curendstr}</b> (instead of <s>{lastendstr}</s>)'
    if end_later:
        return f'was extended to <b>{curendstr}</b> (instead of <s>{lastendstr}</s>)'


def _quick_diff(left, right):
    left = re.sub(r'https?://\S+', '', left)
    left = re.sub(r'\s+', ' ', left).strip()
    right = re.sub(r'https?://\S+', '', right)
    right = re.sub(r'\s+', ' ', right).strip()
    if left == right:
        return
    if not left:
        left = '(empty)'
    if not right:
        right = '(empty)'
    i = 0
    while i < len(left) and i < len(right) and right[i] == left[i]:
        i += 1
    if i > 10:
        left = '\u2026' + left[i - 9:]
        right = '\u2026' + right[i - 9:]
    if len(left) > 40:
        left = left[:39] + '\u2026'
    if len(right) > 40:
        right = right[:39] + '\u2026'
    return left, right


def _generate_preamble(preamble, events):
    if not events:
        return preamble or 'No upcoming events!'

    if preamble and preamble.endswith(':') and 'events' in preamble.lower():
        return preamble

    if len(events) == 1:
        text = "There's an event coming up:"
    elif len(events) == 2:
        text = 'There are a couple events coming up:'
    elif len(events) == 3:
        text = 'There are a few events coming up:'
    else:
        text = 'There are a bunch of events coming up:'
    if preamble:
        if preamble[-1] in ('.', '?', '!'):
            preamble += ' '
        else:
            preamble += ' \u2022 '
        text = text[0].lower() + text[1:]
        if 'event' in preamble.lower():
            text = 'Speaking of which, ' + text
        else:
            text = 'Also, ' + text
        text = preamble + text
    return text



"""Tests for metabot.modules.natlang."""

import pytest

from metabot.calendars import loader
from metabot.modules import natlang


@pytest.fixture
def conversation(build_conversation):  # pylint: disable=missing-docstring
    return build_conversation(natlang)


# pylint: disable=line-too-long


def test_simple(conversation, monkeypatch):  # pylint: disable=redefined-outer-name
    """Verify the natlang module."""

    assert conversation.message('When is the next sock hop?') == ''

    monkeypatch.setattr('time.time', lambda: 1000.)

    other_cal = loader.get('static:other_cal')
    assert other_cal.calcode == '2a366983'
    other_cal.events = {
        '2a366983:alpha': {
            'description': 'Alpha Description',
            'end': 3000,
            'local_id': '2a366983:alpha',
            'location': 'Alpha Venue, Rest of Alpha Location',
            'start': 2000,
            'summary': 'Other Sock Hop',
        },
    }
    conversation.multibot.multical.add('static:other_cal')
    conversation.multibot.calendars['2a366983'] = {'name': 'Other Group'}

    my_cal = loader.get('static:my_cal')
    assert my_cal.calcode == 'fc2d5e05'
    my_cal.events = {
        'fc2d5e05:bravo': {
            'description': '',
            'end': 5000,
            'local_id': 'fc2d5e05:bravo',
            'location': 'My Venue, Rest of My Location',
            'start': 4000,
            'summary': 'My Sock Hop',
        },
        'fc2d5e05:charlie': {
            'description': '',
            'end': 7000,
            'local_id': 'fc2d5e05:charlie',
            'location': 'My Venue, Rest of My Location',
            'start': 6000,
            'summary': 'Special Sock Hop',
        },
        'fc2d5e05:delta': {
            'description': '',
            'end': 9000,
            'local_id': 'fc2d5e05:delta',
            'location': 'Alternate Venue, Rest of Special Location',
            'start': 8000,
            'summary': 'Alternate Sock Hop',
        },
    }
    conversation.multibot.multical.add('static:my_cal')
    conversation.multibot.calendars['fc2d5e05'] = {'name': 'My Group'}

    conversation.bot.config['issue37']['events']['users']['1000']['calendars'] = 'fc2d5e05'

    assert conversation.message('When is the next sock hop?') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
<i>The next sock hop I know about is:</i>

<b>Other Sock Hop</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyAyYTM2Njk4MzphbHBoYSBVVEM">🔜 ¹⁶ ᵐⁱⁿ Thu 1ˢᵗ, 12:33–12:50ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>

<i>The next sock hop on the My Group calendar is:</i>

<b>My Sock Hop</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyBmYzJkNWUwNTpicmF2byBVVEM">🔜 ⁵⁰ ᵐⁱⁿ Thu 1ˢᵗ, 1:06–1:23ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=My+Venue%2C+Rest+of+My+Location">My Venue</a>

<i>The next Special Sock Hop is:</i>

<b>Special Sock Hop</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyBmYzJkNWUwNTpjaGFybGllIFVUQw">🔜 ¹ʰ²³ᵐ Thu 1ˢᵗ, 1:40–1:56ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=My+Venue%2C+Rest+of+My+Location">My Venue</a>

<i>The next sock hop at Alternate Venue is:</i>

<b>Alternate Sock Hop</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyBmYzJkNWUwNTpkZWx0YSBVVEM">🔜 ¹ʰ⁵⁶ᵐ Thu 1ˢᵗ, 2:13–2:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alternate+Venue%2C+Rest+of+Special+Location">Alternate Venue</a>
"""

    assert conversation.message('When is the next my sock hop?') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
<i>The next my sock hop is:</i>

<b>My Sock Hop</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyBmYzJkNWUwNTpicmF2byBVVEM">🔜 ⁵⁰ ᵐⁱⁿ Thu 1ˢᵗ, 1:06–1:23ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=My+Venue%2C+Rest+of+My+Location">My Venue</a>
"""



"""Quick and dirty ISO 8601 -> timestamp converter."""

import datetime
import re

import pytz


def totimestamp(isoformat, timezone=None):
    """Convert a string like YYYY-MM-DDTHH:MM:SS+HH:MM to the number of seconds since 1970-01-01."""

    year, month, day, _, hour, minute, second, fraction, offset, utchour, utcminute = re.match(
        r'(\d{4})-?(\d{1,2})-?(\d{1,2})'
        r'(T(\d{1,2}):?(\d{1,2}):?(\d{1,2})'
        r'([.]\d+)?'
        r'(([-+]\d{1,2}):?(\d{1,2})|Z)?'
        r')?$', isoformat).groups()
    dtime = datetime.datetime(int(year), int(month), int(day), int(hour or 0), int(minute or 0),
                              int(second or 0), int(1000000 * float(fraction or 0)))
    if utchour and utcminute:
        tzinfo = _TimeZone(int(utchour) + float(utcminute) / 60)
    elif offset == 'Z':
        tzinfo = pytz.utc
    elif timezone:
        tzinfo = pytz.timezone(timezone)
    else:
        tzinfo = pytz.utc
    return (tzinfo.localize(dtime) - _EPOCH).total_seconds()


class _TimeZone(datetime.tzinfo):
    __slots__ = ('_utcoffset',)

    def __init__(self, utcoffset):
        super().__init__()
        self._utcoffset = datetime.timedelta(hours=utcoffset)

    @staticmethod
    def tzname(unused_dt):  # pragma: no cover  pylint: disable=arguments-differ
        return 'x'

    def utcoffset(self, unused_dt):
        return self._utcoffset

    @staticmethod
    def dst(unused_dt):  # pragma: no cover  pylint: disable=arguments-differ
        return datetime.timedelta(0)

    def localize(self, dt):
        """See https://pypi.org/project/pytz/#example-usage."""

        return dt.replace(tzinfo=self)


_EPOCH = datetime.datetime(1970, 1, 1, tzinfo=pytz.utc)



"""Set up new bots."""

import logging

import ntelebot


def modhelp(unused_ctx, unused_modconf, sections):  # pylint: disable=missing-docstring
    sections['commands'].add('/newbot \u2013 Set up a new bot')


def moddispatch(ctx, msg, unused_modconf):  # pylint: disable=missing-docstring
    if ctx.type in ('message', 'callback_query') and ctx.command == 'newbot':
        return default(ctx, msg)

    return False


def default(ctx, msg):  # pylint: disable=missing-docstring
    ctx.private = True
    msg.path('/newbot', 'Add a bot')
    msg.action = 'Paste a bot API Token'

    token, _ = ctx.split(2)

    if token:
        try:
            bot = ntelebot.bot.Bot(token)
        except AssertionError:
            msg.add("Oops, <code>%s</code> doesn't look like an API Token.", token)
            token = None

    if not token:
        msg.add("To create a new bot, let me know your bot account's API Token. This is a code "
                'that looks like:')
        msg.add('<pre>123456:ABC-DEF1234ghIkl-zyx57W2v1u123ew11</pre>')
        msg.add('If you are creating a bot entirely from scratch, open a private chat with '
                '@BotFather and follow the process at '
                'https://core.telegram.org/bots#creating-a-new-bot to create a new bot account. At '
                'the end of that process, you will receive an API Token that you can paste here.')
        msg.add('Otherwise, open a private chat with @BotFather, type <code>/mybots</code>, select '
                'the bot account you want to use, select <code>API Token</code>, then copy the '
                'code and paste it here:')
        return

    try:
        bot_info = bot.get_me()  # pylint: disable=possibly-used-before-assignment
    except ntelebot.errors.Unauthorized:
        return msg.add(
            'Woops, Telegram told me <code>%s</code> is unauthorized, meaning the code is either '
            'incomplete or out of date. If you generated it yourself, open a private chat with '
            "@BotFather, type <code>/mybots</code>, select the bot account you're trying to use, "
            'select <code>API Token</code>, then copy the code and paste it here. If you got the '
            'code from someone else, send them these instructions (including the token I used). If '
            "the code you got from BotFather isn't working, select <code>Revoke current "
            'token</code> to generate a new one, then paste that new one here:', token)
    except ntelebot.errors.Error as exc:  # pragma: no cover
        return msg.add(
            'Woops, while trying to use <code>%s</code> I got error %s (<code>%s</code>).', token,
            exc.error_code, exc.description)

    ctx.reply_html('Cool, that API Token is for <code>%s</code>. Give me another moment...',
                   bot_info['username'])

    try:
        bot.get_updates(limit=1, timeout=10)
    except ntelebot.errors.Conflict:
        return msg.add(
            'Woops, it looks like this bot account is already in use. Make sure no other bot '
            'programs are running using this API Token and paste the token again, or use another '
            'one:')
    except ntelebot.errors.Error as exc:
        logging.exception('While polling %r:', token)
        return msg.add(
            'Woops, while trying to use <code>%s</code> I got error %s (<code>%s</code>).', token,
            exc.error_code, exc.description)

    username = ctx.bot.multibot.add_bot(token)
    ctx.bot.multibot.conf['bots'][username]['issue37']['admin']['admins'] = [ctx.user['id']]
    ctx.bot.multibot.run_bot(username)

    msg.action = 'Configure your new bot'
    msg.add(
        'Yay, I am now running <code>%s</code> (<code>%s</code>). Open a private chat with @%s and '
        'type <code>/admin</code> to continue setup.', bot_info['username'], bot_info['first_name'],
        bot_info['username'])



"""Quick wrapper around pytz's zone.tab file (see https://bugs.launchpad.net/pytz/+bug/1324972)."""

import collections
import re

import pytz


def _zone_tab():
    zonetype = collections.namedtuple('Zone', ('code', 'coordinates', 'zone', 'comment'))
    with pytz.open_resource('zone.tab') as fobj:
        for line in fobj:
            line = line.decode('UTF-8')
            if line.startswith('#'):
                continue
            pieces = line.rstrip().split(None, 3)
            code, coordinates, zone = pieces[:3]
            comment = len(pieces) == 4 and pieces[3] or ''
            if zone not in pytz.all_timezones_set:
                continue

            coordinates = _parse_iso6709(coordinates)
            if not coordinates:
                continue
            lat, lon = coordinates

            yield zonetype(code, (lat, lon), zone, comment)


ZONE_TAB = pytz.LazyList(_zone_tab())


def _parse_iso6709(iso6709):
    match = re.match('^([-+][0-9]+)([-+][0-9]+)$', iso6709)
    if not match:
        return
    return map(_parse_iso6709_coord, match.groups())


def _parse_iso6709_coord(coord):
    if len(coord) % 2:
        deg = float(coord[:3])
        coord = coord[3:]
    else:
        deg = float(coord[:4])
        coord = coord[4:]
    div = 60
    while coord:
        deg += float(coord[:2]) / div
        div *= 60
        coord = coord[2:]
    return deg



"""A mutation-logging, key-implying dict (with accompanying list)."""


class ImplicitTrackingDict(dict):
    """A mutation-logging, key-implying dict."""

    def __init__(self, value=None, log=None, path=(), **kwargs):
        super().__init__(**kwargs)
        if log is not None:
            self.log = log
        else:
            self.log = {}
        self.path = path
        if value is not None:
            self.update(value)

    def finalize(self):
        """Trim empty hanging dicts, clear the audit log, and return a copy of the log."""

        self._trim()
        log = dict(self.log)
        self.log.clear()
        return log

    def _trim(self):
        for key, value in list(self.items()):
            if isinstance(value, ImplicitTrackingDict):
                value._trim()  # pylint: disable=protected-access
                if not value:
                    self.pop(key)

    def audit(self, path, value, current):
        """Record a mutation of root-path path from current to value."""

        if path not in self.log:
            self.log[path] = (value, current)
        else:
            orig = self.log[path][1]
            if value == orig:
                self.log.pop(path)
            else:
                self.log[path] = (value, orig)

    def clear(self):
        for key in list(self):
            self.pop(key)

    def pop(self, key, default=None):
        path = self.path + (key,)
        current = super().pop(key, None)
        if current is None:
            return default
        if isinstance(current, ImplicitTrackingDict):
            ret = dict(current)
            current.clear()
            current = ret
        else:
            self.audit(path, None, current)
        return current

    def update(self, values):
        for key, value in values.items():
            self[key] = value

    def __delitem__(self, key):
        self.pop(key)

    def __missing__(self, key):
        self[key] = {}
        return self[key]

    def __setitem__(self, key, value):
        if value is None:
            return self.__delitem__(key)

        current = self.get(key)
        if value == current:
            return
        if isinstance(current, (ImplicitTrackingDict, TrackingList)):
            current.clear()
            current = None

        path = self.path + (key,)
        if isinstance(value, dict):
            value = ImplicitTrackingDict(value=value, log=self.log, path=path)
        elif isinstance(value, (list, tuple)):
            value = TrackingList(value=value, log=self.log, path=path)
        elif isinstance(value, bytes):  # PyYAML converts ASCII strings to str in Python 2.7.
            value = value.decode('ascii')
        else:
            assert isinstance(value, (int, str)), repr(value)

        super().__setitem__(key, value)

        if not isinstance(value, (ImplicitTrackingDict, TrackingList)):
            self.audit(path, value, current)


class TrackingList(list):
    """A mutation-logging list."""

    def __init__(self, value=None, log=None, path=(), **kwargs):
        super().__init__(**kwargs)
        if log is not None:
            self.log = log
        else:
            self.log = {}  # pragma: no cover
        self.path = path
        if value is not None:
            self.extend(value)

    def audit(self, current):
        """Record a mutation of root-path self.path from current to self."""

        value = tuple(self)
        if self.path not in self.log:
            self.log[self.path] = (value, current)
        else:
            orig = self.log[self.path][1]
            if value == orig:
                self.log.pop(self.path)
            else:
                self.log[self.path] = (value, orig)

    def append(self, value):
        assert isinstance(value, (int, str)), value
        current = tuple(self)
        super().append(value)
        self.audit(current)

    def clear(self):  # pylint: disable=missing-docstring
        while self:
            self.pop()

    def extend(self, values):
        for value in values:
            self.append(value)

    def insert(self, index, value):
        assert isinstance(value, (int, str)), value
        current = tuple(self)
        super().insert(index, value)
        self.audit(current)

    def pop(self, index=-1):
        current = tuple(self)
        ret = super().pop(index)
        self.audit(current)
        return ret

    def remove(self, value):
        current = tuple(self)
        super().remove(value)
        self.audit(current)

    def reverse(self):
        current = tuple(self)
        super().reverse()
        self.audit(current)

    def sort(self, **kwargs):
        current = tuple(self)
        super().sort(**kwargs)
        self.audit(current)

    def __setitem__(self, index, value):
        assert isinstance(value, (int, str)), value
        if self[index] == value:
            return
        current = tuple(self)
        super().__setitem__(index, value)
        self.audit(current)



"""Telegram-specific HTML sanitizer."""

import html
import html.parser
import logging


def escape(text):  # pylint: disable=missing-docstring
    return text.replace('&', '&amp;').replace('<', '&lt;').replace('>',
                                                                   '&gt;').replace('"', '&quot;')


class _HTMLSanitizer(html.parser.HTMLParser):
    __pieces = __strip = __remaining = __stack = None

    def sanitize(self, text, strip=False, length=None):
        """Sanitize (or optionally entirely strip) HTML for use in Telegram."""

        self.__pieces = []
        self.__strip = strip
        self.__remaining = length
        self.__stack = []
        self.feed(text)
        self.close()
        self.handle_endtag(None)
        return ''.join(self.__pieces)

    naked = {'b', 'del', 'em', 'i', 'ins', 'pre', 's', 'strike', 'strong', 'tg-spoiler', 'u'}
    coupled = {
        'a': 'href',
        'blockquote': 'expandable',
        'code': 'class',
        'span': 'class',
        'tg-emoji': 'emoji-id',
    }

    def handle_starttag(self, tag, attrs):  # pylint: disable=too-many-branches,too-many-return-statements
        if self.__remaining is not None and not self.__remaining:
            return
        if tag in ('br', 'div', 'p'):
            self.__append('\n')
            return
        if self.__strip:
            return
        if tag in self.naked:
            self.__stack.append(tag)
            self.__pieces.append(f'<{tag}>')
            return
        if tag not in self.coupled:
            return

        attrs = dict(attrs)
        attribute_name = self.coupled[tag]
        attribute_value = attrs.get(attribute_name, '')

        # pylint: disable=line-too-long
        # if (tag_name == "a" && attribute_name == Slice("href")) {
        #   argument = std::move(attribute_value);
        # } else if (tag_name == "code" && attribute_name == Slice("class") &&
        #            begins_with(attribute_value, "language-")) {
        #   argument = attribute_value.substr(9);
        # } else if (tag_name == "span" && attribute_name == Slice("class") && begins_with(attribute_value, "tg-")) {
        #   argument = attribute_value.substr(3);
        # } else if (tag_name == "tg-emoji" && attribute_name == Slice("emoji-id")) {
        #   argument = std::move(attribute_value);
        # } else if (tag_name == "blockquote" && attribute_name == Slice("expandable")) {
        #   argument = "1";
        # }
        argument = None
        if tag == 'a':
            argument = attribute_value
        elif tag == 'code' and attribute_value.startswith('language-'):
            argument = attribute_value[9:]
        elif tag == 'span' and attribute_value.startswith('tg-'):
            argument = attribute_value[3:]
        elif tag == 'tg-emoji':
            argument = attribute_value
        elif tag == 'blockquote' and 'expandable' in attrs:
            argument = '1'

        # if (tag_name == "span" && argument != "spoiler") {
        #   return Status::Error(400, PSLICE()
        #                                 << "Tag \"span\" must have class \"tg-spoiler\" at byte offset " << begin_pos);
        # }
        if tag == 'span' and argument != 'spoiler':
            return

        # } else if (tag_name == "tg-emoji") {
        #   auto r_document_id = to_integer_safe<int64>(nested_entities.back().argument);
        #   if (r_document_id.is_error() || r_document_id.ok() == 0) {
        #     return Status::Error(400, "Invalid custom emoji identifier specified");
        #   }
        if tag == 'tg-emoji' and (not argument.isdigit() or not 0 < int(argument) < 2**63):
            return

        if tag == 'a' and not argument:
            return

        self.__stack.append(tag)
        if not argument:
            self.__pieces.append(f'<{tag}>')
        elif not attribute_value:
            self.__pieces.append(f'<{tag} {attribute_name}>')
        else:
            self.__pieces.append(f'<{tag} {attribute_name}="{escape(attribute_value)}">')

    def handle_endtag(self, tag):
        if tag and tag not in self.__stack:
            return

        while self.__stack:
            lasttag = self.__stack.pop()
            self.__pieces.append(f'</{lasttag}>')
            if lasttag == tag:
                break

    def __append(self, data):
        if self.__remaining is None:
            return self.__pieces.append(data)

        if len(data) <= self.__remaining and '&' not in data:
            self.__remaining -= len(data)
            return self.__pieces.append(data)

        inentity = False
        for char in data:
            if not self.__remaining:
                break
            self.__pieces.append(char)
            if inentity:
                if char == ';':
                    self.__remaining -= 1
                    inentity = False
            else:
                if char == '&':
                    inentity = True
                else:
                    self.__remaining -= 1

    def handle_data(self, data):
        self.__append(escape(data))

    def error(self, message):  # pragma: no cover  pylint: disable=missing-function-docstring
        # Remove when support for Python below 3.10 is dropped.
        logging.error('HTML sanitizer error: %s', message)


def sanitize(text, strip=False):
    """Sanitize (or optionally entirely strip) HTML for use in Telegram."""

    return _HTMLSanitizer().sanitize(text, strip=strip)


def truncate(text, length):
    """Truncate text to at most length non-markup characters."""

    return _HTMLSanitizer().sanitize(text, length=length)



"""Tests for metabot.util.adminui."""

import pytest

from metabot.modules import echo
from metabot.modules import moderator


@pytest.fixture
def conversation(build_conversation):  # pylint: disable=missing-docstring
    conv = build_conversation(echo, moderator)
    conv.groupconf = conv.bot.config['issue37']['moderator']['-1001000001000']
    conv.groupconf['title'] = 'Test Group'
    return conv


# pylint: disable=line-too-long


def test_bool(conversation):  # pylint: disable=redefined-outer-name
    """Test adminui.bool."""

    echoconf = conversation.bot.config['issue37']['echo']['dummy']

    assert 'paginate' not in echoconf

    assert conversation.message('/admin modulestestbot echo dummy paginate') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › echo › dummy: <b>Choose a field</b>

Enabled <code>paginate</code>.
[paginate (yes) • For multiline messages, display just one line at a time? | /admin modulestestbot echo dummy paginate]
[private (no) • Send the message in group chats, or just in private? | /admin modulestestbot echo dummy private]
[text • The message, sticker, or image to send in response to /dummy. | /admin modulestestbot echo dummy text]
[Back | /admin modulestestbot echo]
"""

    assert echoconf['paginate'] is True

    assert conversation.message('/admin modulestestbot echo dummy paginate') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › echo › dummy: <b>Choose a field</b>

Disabled <code>paginate</code>.
[paginate (no) • For multiline messages, display just one line at a time? | /admin modulestestbot echo dummy paginate]
[private (no) • Send the message in group chats, or just in private? | /admin modulestestbot echo dummy private]
[text • The message, sticker, or image to send in response to /dummy. | /admin modulestestbot echo dummy text]
[Back | /admin modulestestbot echo]
"""

    assert 'paginate' not in echoconf


def test_daysofweek(conversation):  # pylint: disable=redefined-outer-name
    """Test adminui.daysofweek."""

    assert conversation.message('/admin modulestestbot moderator -1001000001000 daily dow') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › daily › dow: <b>Select a day of the week to toggle</b>

Which days of the week should I announce upcoming events on?

All days are currently <b>enabled</b>.

Select a day of the week to toggle:
[disable Sunday | /admin modulestestbot moderator -1001000001000 daily dow 6] [disable Monday | /admin modulestestbot moderator -1001000001000 daily dow 0]
[disable Tuesday | /admin modulestestbot moderator -1001000001000 daily dow 1] [disable Wednesday | /admin modulestestbot moderator -1001000001000 daily dow 2]
[disable Thursday | /admin modulestestbot moderator -1001000001000 daily dow 3] [disable Friday | /admin modulestestbot moderator -1001000001000 daily dow 4]
[disable Saturday | /admin modulestestbot moderator -1001000001000 daily dow 5] [disable every day | /admin modulestestbot moderator -1001000001000 daily dow none]
[Back | /admin modulestestbot moderator -1001000001000 daily]
"""

    assert 'daily' not in conversation.groupconf

    assert conversation.message('none') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › daily › dow: <b>Select a day of the week to toggle</b>

Which days of the week should I announce upcoming events on?

All days are currently <b>disabled</b>.

Select a day of the week to toggle:
[enable Sunday | /admin modulestestbot moderator -1001000001000 daily dow 6] [enable Monday | /admin modulestestbot moderator -1001000001000 daily dow 0]
[enable Tuesday | /admin modulestestbot moderator -1001000001000 daily dow 1] [enable Wednesday | /admin modulestestbot moderator -1001000001000 daily dow 2]
[enable Thursday | /admin modulestestbot moderator -1001000001000 daily dow 3] [enable Friday | /admin modulestestbot moderator -1001000001000 daily dow 4]
[enable Saturday | /admin modulestestbot moderator -1001000001000 daily dow 5] [enable every day | /admin modulestestbot moderator -1001000001000 daily dow all]
[Back | /admin modulestestbot moderator -1001000001000 daily]
"""

    assert conversation.groupconf['daily'] == {'dow': 127}

    assert conversation.message('1') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › daily › dow: <b>Select a day of the week to toggle</b>

Which days of the week should I announce upcoming events on?

Enabled for Tuesday.

Select a day of the week to toggle:
[enable Sunday | /admin modulestestbot moderator -1001000001000 daily dow 6] [enable Monday | /admin modulestestbot moderator -1001000001000 daily dow 0]
[disable Tuesday | /admin modulestestbot moderator -1001000001000 daily dow 1] [enable Wednesday | /admin modulestestbot moderator -1001000001000 daily dow 2]
[enable Thursday | /admin modulestestbot moderator -1001000001000 daily dow 3] [enable Friday | /admin modulestestbot moderator -1001000001000 daily dow 4]
[enable Saturday | /admin modulestestbot moderator -1001000001000 daily dow 5] [enable every day | /admin modulestestbot moderator -1001000001000 daily dow all]
[Back | /admin modulestestbot moderator -1001000001000 daily]
"""

    assert conversation.groupconf['daily'] == {'dow': 125}

    assert conversation.message('6') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › daily › dow: <b>Select a day of the week to toggle</b>

Which days of the week should I announce upcoming events on?

Enabled for Sunday and Tuesday.

Select a day of the week to toggle:
[disable Sunday | /admin modulestestbot moderator -1001000001000 daily dow 6] [enable Monday | /admin modulestestbot moderator -1001000001000 daily dow 0]
[disable Tuesday | /admin modulestestbot moderator -1001000001000 daily dow 1] [enable Wednesday | /admin modulestestbot moderator -1001000001000 daily dow 2]
[enable Thursday | /admin modulestestbot moderator -1001000001000 daily dow 3] [enable Friday | /admin modulestestbot moderator -1001000001000 daily dow 4]
[enable Saturday | /admin modulestestbot moderator -1001000001000 daily dow 5] [enable every day | /admin modulestestbot moderator -1001000001000 daily dow all]
[Back | /admin modulestestbot moderator -1001000001000 daily]
"""

    assert conversation.groupconf['daily'] == {'dow': 61}

    assert conversation.message('3') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › daily › dow: <b>Select a day of the week to toggle</b>

Which days of the week should I announce upcoming events on?

Enabled for Sunday, Tuesday, and Thursday.

Select a day of the week to toggle:
[disable Sunday | /admin modulestestbot moderator -1001000001000 daily dow 6] [enable Monday | /admin modulestestbot moderator -1001000001000 daily dow 0]
[disable Tuesday | /admin modulestestbot moderator -1001000001000 daily dow 1] [enable Wednesday | /admin modulestestbot moderator -1001000001000 daily dow 2]
[disable Thursday | /admin modulestestbot moderator -1001000001000 daily dow 3] [enable Friday | /admin modulestestbot moderator -1001000001000 daily dow 4]
[enable Saturday | /admin modulestestbot moderator -1001000001000 daily dow 5] [enable every day | /admin modulestestbot moderator -1001000001000 daily dow all]
[Back | /admin modulestestbot moderator -1001000001000 daily]
"""

    assert conversation.groupconf['daily'] == {'dow': 53}

    assert conversation.message('all') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › daily › dow: <b>Select a day of the week to toggle</b>

Which days of the week should I announce upcoming events on?

All days are currently <b>enabled</b>.

Select a day of the week to toggle:
[disable Sunday | /admin modulestestbot moderator -1001000001000 daily dow 6] [disable Monday | /admin modulestestbot moderator -1001000001000 daily dow 0]
[disable Tuesday | /admin modulestestbot moderator -1001000001000 daily dow 1] [disable Wednesday | /admin modulestestbot moderator -1001000001000 daily dow 2]
[disable Thursday | /admin modulestestbot moderator -1001000001000 daily dow 3] [disable Friday | /admin modulestestbot moderator -1001000001000 daily dow 4]
[disable Saturday | /admin modulestestbot moderator -1001000001000 daily dow 5] [disable every day | /admin modulestestbot moderator -1001000001000 daily dow none]
[Back | /admin modulestestbot moderator -1001000001000 daily]
"""

    assert 'daily' not in conversation.groupconf


def test_forward(conversation):  # pylint: disable=redefined-outer-name
    """Test adminui.forward."""

    conversation.bot.config['issue37']['moderator']['-1002000002000']['title'] = 'Forward Source'

    assert conversation.message('/admin modulestestbot moderator -1001000001000') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000: <b>Choose a field</b>
[calendars • Which calendars should be listed in /events? | /admin modulestestbot moderator -1001000001000 calendars]
[daily • Should I announce upcoming events once a day? | /admin modulestestbot moderator -1001000001000 daily]
[forward • Automatically forward messages from one chat to this one. | /admin modulestestbot moderator -1001000001000 forward]
[greeting • How should I greet people when they join? | /admin modulestestbot moderator -1001000001000 greeting]
[maxeventscount • How many events should be listed in /events? | /admin modulestestbot moderator -1001000001000 maxeventscount]
[maxeventsdays • How many days into the future should /events look? | /admin modulestestbot moderator -1001000001000 maxeventsdays]
[timezone • What time zone should be used in /events? | /admin modulestestbot moderator -1001000001000 timezone]
[Back | /admin modulestestbot moderator]
"""

    assert conversation.message('/admin modulestestbot moderator -1001000001000 forward') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › forward: <b>Choose a field</b>

Automatically forward messages from one chat to this one.
[from • What group should messages be forwarded from? | /admin modulestestbot moderator -1001000001000 forward from]
[notify (no) • Should forwarded messages trigger a notification? | /admin modulestestbot moderator -1001000001000 forward notify]
[Back | /admin modulestestbot moderator -1001000001000]
"""

    assert conversation.message(
        '/admin modulestestbot moderator -1001000001000 forward from') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › forward › from: <b>Select a group</b>

What group should messages be forwarded from?

Select a group:
[-1001000001000 (Test Group) | /admin modulestestbot moderator -1001000001000 forward from -1001000001000]
[-1002000002000 (Forward Source) | /admin modulestestbot moderator -1001000001000 forward from -1002000002000]
[Back | /admin modulestestbot moderator -1001000001000 forward]
"""

    assert conversation.message('-1002000002000') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › forward: <b>Choose a field</b>

Set <code>from</code> to <code>-1002000002000</code>.

Automatically forward messages from one chat to this one.
[from (-10020000…) • What group should messages be forwarded from? | /admin modulestestbot moderator -1001000001000 forward from]
[notify (no) • Should forwarded messages trigger a notification? | /admin modulestestbot moderator -1001000001000 forward notify]
[Back | /admin modulestestbot moderator -1001000001000]
"""

    assert conversation.message('/admin modulestestbot moderator -1001000001000') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000: <b>Choose a field</b>
[calendars • Which calendars should be listed in /events? | /admin modulestestbot moderator -1001000001000 calendars]
[daily • Should I announce upcoming events once a day? | /admin modulestestbot moderator -1001000001000 daily]
[forward • Automatically forward messages from one chat to this one. | /admin modulestestbot moderator -1001000001000 forward]
[greeting • How should I greet people when they join? | /admin modulestestbot moderator -1001000001000 greeting]
[maxeventscount • How many events should be listed in /events? | /admin modulestestbot moderator -1001000001000 maxeventscount]
[maxeventsdays • How many days into the future should /events look? | /admin modulestestbot moderator -1001000001000 maxeventsdays]
[timezone • What time zone should be used in /events? | /admin modulestestbot moderator -1001000001000 timezone]
[Back | /admin modulestestbot moderator]
"""


def test_integer(conversation):  # pylint: disable=redefined-outer-name
    """Test adminui.integer."""

    assert conversation.message('/admin modulestestbot moderator -1001000001000 daily') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › daily: <b>Choose a field</b>

Should I announce upcoming events once a day?
[dow • Which days of the week should I announce upcoming events on? | /admin modulestestbot moderator -1001000001000 daily dow]
[hour • At what hour? | /admin modulestestbot moderator -1001000001000 daily hour]
[pin (no) • Should the announcement be pinned? | /admin modulestestbot moderator -1001000001000 daily pin]
[text • One or more messages (one per line) to use/cycle through for the daily announcement. | /admin modulestestbot moderator -1001000001000 daily text]
[Back | /admin modulestestbot moderator -1001000001000]
"""

    assert conversation.message('/admin modulestestbot moderator -1001000001000 daily hour') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › daily › hour: <b>Type a new value for hour</b>

At what hour?

Type your new value, or type "off" to disable/reset to default.
[Back | /admin modulestestbot moderator -1001000001000 daily]
"""

    assert 'daily' not in conversation.groupconf

    assert conversation.message('8') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › daily: <b>Choose a field</b>

Set <code>hour</code> to <code>8</code>.

Should I announce upcoming events once a day?
[dow • Which days of the week should I announce upcoming events on? | /admin modulestestbot moderator -1001000001000 daily dow]
[hour (8) • At what hour? | /admin modulestestbot moderator -1001000001000 daily hour]
[pin (no) • Should the announcement be pinned? | /admin modulestestbot moderator -1001000001000 daily pin]
[text • One or more messages (one per line) to use/cycle through for the daily announcement. | /admin modulestestbot moderator -1001000001000 daily text]
[Back | /admin modulestestbot moderator -1001000001000]
"""

    assert conversation.groupconf['daily'] == {'hour': 8}

    assert conversation.message('/admin modulestestbot moderator -1001000001000 daily hour') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › daily › hour: <b>Type a new value for hour</b>

At what hour?

<code>hour</code> is currently <code>8</code>.

Type your new value, or type "off" to disable/reset to default.
[Back | /admin modulestestbot moderator -1001000001000 daily]
"""

    assert conversation.message('9') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › daily: <b>Choose a field</b>

Changed <code>hour</code> from <code>8</code> to <code>9</code>.

Should I announce upcoming events once a day?
[dow • Which days of the week should I announce upcoming events on? | /admin modulestestbot moderator -1001000001000 daily dow]
[hour (9) • At what hour? | /admin modulestestbot moderator -1001000001000 daily hour]
[pin (no) • Should the announcement be pinned? | /admin modulestestbot moderator -1001000001000 daily pin]
[text • One or more messages (one per line) to use/cycle through for the daily announcement. | /admin modulestestbot moderator -1001000001000 daily text]
[Back | /admin modulestestbot moderator -1001000001000]
"""

    assert conversation.groupconf['daily'] == {'hour': 9}

    assert conversation.message('hour off') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › daily: <b>Choose a field</b>

Unset <code>hour</code> (was <code>9</code>).

Should I announce upcoming events once a day?
[dow • Which days of the week should I announce upcoming events on? | /admin modulestestbot moderator -1001000001000 daily dow]
[hour • At what hour? | /admin modulestestbot moderator -1001000001000 daily hour]
[pin (no) • Should the announcement be pinned? | /admin modulestestbot moderator -1001000001000 daily pin]
[text • One or more messages (one per line) to use/cycle through for the daily announcement. | /admin modulestestbot moderator -1001000001000 daily text]
[Back | /admin modulestestbot moderator -1001000001000]
"""

    assert 'daily' not in conversation.groupconf

    assert conversation.message('hour off') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › daily: <b>Choose a field</b>

<code>hour</code> is already unset.

Should I announce upcoming events once a day?
[dow • Which days of the week should I announce upcoming events on? | /admin modulestestbot moderator -1001000001000 daily dow]
[hour • At what hour? | /admin modulestestbot moderator -1001000001000 daily hour]
[pin (no) • Should the announcement be pinned? | /admin modulestestbot moderator -1001000001000 daily pin]
[text • One or more messages (one per line) to use/cycle through for the daily announcement. | /admin modulestestbot moderator -1001000001000 daily text]
[Back | /admin modulestestbot moderator -1001000001000]
"""


def test_timezone(conversation):  # pylint: disable=redefined-outer-name
    """Test adminui.timezone."""

    assert conversation.message('/admin modulestestbot moderator -1001000001000 timezone') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › timezone: <b>Type your 2-letter country code</b>

Type your 2-letter country code (like US, CA, GB, etc.).
[Back | /admin modulestestbot moderator -1001000001000]
"""

    assert conversation.message(
        '/admin modulestestbot moderator -1001000001000 timezone dummy') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › timezone: <b>Type your 2-letter country code</b>

Unknown country code <code>DUMMY</code>.

Type your 2-letter country code (like US, CA, GB, etc.).
[Back | /admin modulestestbot moderator -1001000001000]
"""

    assert conversation.message(
        '/admin modulestestbot moderator -1001000001000 timezone gb') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › timezone: <b>Choose a primary city</b>

What time zone should be used in /events?

Choose a primary city:
[London | /admin modulestestbot moderator -1001000001000 timezone Europe/London]
[Back | /admin modulestestbot moderator -1001000001000]
"""

    assert conversation.message(
        '/admin modulestestbot moderator -1001000001000 timezone us') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › timezone: <b>Choose a primary city</b>

What time zone should be used in /events?

Choose a primary city:
[New York (Eastern (most areas)) | /admin modulestestbot moderator -1001000001000 timezone America/New_York]
[Detroit (Eastern - MI (most areas)) | /admin modulestestbot moderator -1001000001000 timezone America/Detroit]
[Louisville (Eastern - KY (Louisville area)) | /admin modulestestbot moderator -1001000001000 timezone America/Kentucky/Louisville]
[Monticello (Eastern - KY (Wayne)) | /admin modulestestbot moderator -1001000001000 timezone America/Kentucky/Monticello]
[Indianapolis (Eastern - IN (most areas)) | /admin modulestestbot moderator -1001000001000 timezone America/Indiana/Indianapolis]
[Vincennes (Eastern - IN (Da, Du, K, Mn)) | /admin modulestestbot moderator -1001000001000 timezone America/Indiana/Vincennes]
[Winamac (Eastern - IN (Pulaski)) | /admin modulestestbot moderator -1001000001000 timezone America/Indiana/Winamac]
[\xa0 | /stop] [Next | /admin modulestestbot moderator -1001000001000 timezone US 1]
[Back | /admin modulestestbot moderator -1001000001000]
"""

    assert conversation.message(
        '/admin modulestestbot moderator -1001000001000 timezone US 1') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › timezone: <b>Choose a primary city</b>

What time zone should be used in /events?

Choose a primary city:
[Marengo (Eastern - IN (Crawford)) | /admin modulestestbot moderator -1001000001000 timezone America/Indiana/Marengo]
[Petersburg (Eastern - IN (Pike)) | /admin modulestestbot moderator -1001000001000 timezone America/Indiana/Petersburg]
[Vevay (Eastern - IN (Switzerland)) | /admin modulestestbot moderator -1001000001000 timezone America/Indiana/Vevay]
[Chicago (Central (most areas)) | /admin modulestestbot moderator -1001000001000 timezone America/Chicago]
[Tell City (Central - IN (Perry)) | /admin modulestestbot moderator -1001000001000 timezone America/Indiana/Tell_City]
[Knox (Central - IN (Starke)) | /admin modulestestbot moderator -1001000001000 timezone America/Indiana/Knox]
[Menominee (Central - MI (Wisconsin border)) | /admin modulestestbot moderator -1001000001000 timezone America/Menominee]
[Prev | /admin modulestestbot moderator -1001000001000 timezone US 0] [Next | /admin modulestestbot moderator -1001000001000 timezone US 2]
[Back | /admin modulestestbot moderator -1001000001000]
"""

    assert conversation.message(
        '/admin modulestestbot moderator -1001000001000 timezone US 2') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › timezone: <b>Choose a primary city</b>

What time zone should be used in /events?

Choose a primary city:
[Center (Central - ND (Oliver)) | /admin modulestestbot moderator -1001000001000 timezone America/North_Dakota/Center]
[New Salem (Central - ND (Morton rural)) | /admin modulestestbot moderator -1001000001000 timezone America/North_Dakota/New_Salem]
[Beulah (Central - ND (Mercer)) | /admin modulestestbot moderator -1001000001000 timezone America/North_Dakota/Beulah]
[Denver (Mountain (most areas)) | /admin modulestestbot moderator -1001000001000 timezone America/Denver]
[Boise (Mountain - ID (south), OR (east)) | /admin modulestestbot moderator -1001000001000 timezone America/Boise]
[Phoenix (MST - AZ (except Navajo)) | /admin modulestestbot moderator -1001000001000 timezone America/Phoenix]
[Los Angeles (Pacific) | /admin modulestestbot moderator -1001000001000 timezone America/Los_Angeles]
[Prev | /admin modulestestbot moderator -1001000001000 timezone US 1] [Next | /admin modulestestbot moderator -1001000001000 timezone US 3]
[Back | /admin modulestestbot moderator -1001000001000]
"""

    assert conversation.message(
        '/admin modulestestbot moderator -1001000001000 timezone US 3') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › timezone: <b>Choose a primary city</b>

What time zone should be used in /events?

Choose a primary city:
[Anchorage (Alaska (most areas)) | /admin modulestestbot moderator -1001000001000 timezone America/Anchorage]
[Juneau (Alaska - Juneau area) | /admin modulestestbot moderator -1001000001000 timezone America/Juneau]
[Sitka (Alaska - Sitka area) | /admin modulestestbot moderator -1001000001000 timezone America/Sitka]
[Metlakatla (Alaska - Annette Island) | /admin modulestestbot moderator -1001000001000 timezone America/Metlakatla]
[Yakutat (Alaska - Yakutat) | /admin modulestestbot moderator -1001000001000 timezone America/Yakutat]
[Nome (Alaska (west)) | /admin modulestestbot moderator -1001000001000 timezone America/Nome]
[Adak (Alaska - western Aleutians) | /admin modulestestbot moderator -1001000001000 timezone America/Adak]
[Prev | /admin modulestestbot moderator -1001000001000 timezone US 2] [Next | /admin modulestestbot moderator -1001000001000 timezone US 4]
[Back | /admin modulestestbot moderator -1001000001000]
"""

    assert conversation.message(
        '/admin modulestestbot moderator -1001000001000 timezone US 4') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › timezone: <b>Choose a primary city</b>

What time zone should be used in /events?

Choose a primary city:
[Honolulu (Hawaii) | /admin modulestestbot moderator -1001000001000 timezone Pacific/Honolulu]
[Prev | /admin modulestestbot moderator -1001000001000 timezone US 3] [\xa0 | /stop]
[Back | /admin modulestestbot moderator -1001000001000]
"""

    assert conversation.message(
        '/admin modulestestbot moderator -1001000001000 timezone America/Los_Angeles') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000: <b>Choose a field</b>

Set <code>timezone</code> to <code>America/Los_Angeles</code>.
[calendars • Which calendars should be listed in /events? | /admin modulestestbot moderator -1001000001000 calendars]
[daily • Should I announce upcoming events once a day? | /admin modulestestbot moderator -1001000001000 daily]
[forward • Automatically forward messages from one chat to this one. | /admin modulestestbot moderator -1001000001000 forward]
[greeting • How should I greet people when they join? | /admin modulestestbot moderator -1001000001000 greeting]
[maxeventscount • How many events should be listed in /events? | /admin modulestestbot moderator -1001000001000 maxeventscount]
[maxeventsdays • How many days into the future should /events look? | /admin modulestestbot moderator -1001000001000 maxeventsdays]
[timezone (America/L…) • What time zone should be used in /events? | /admin modulestestbot moderator -1001000001000 timezone]
[Back | /admin modulestestbot moderator]
"""
